#include <vector>
#include <iomanip>
#include <algorithm>
#include <list>
#include <set>
#include <unordered_set>

#include "MeshLoader.h"

int MeshLoader::getLastNodesId() {
    return nodes.rbegin()->id;
}

void MeshLoader::PrintData() {
    std::cout << Nodes() << std::endl
        << FiniteElements() << std::endl
        << BoundaryFiniteElements() << std::endl;
}

void MeshLoader::TransformElementsToQuadr() {
    
    std::unordered_set<Rib, MyHash> ribs;
    for (auto& el : finite_elements) {
        std::vector<int> cur_nodes(el.nodes_id);

        for (auto i = cur_nodes.begin(); i < cur_nodes.end(); ++i) {

            for (auto j = i + 1; j < cur_nodes.end(); ++j) {
                Rib cur_rib(*i, *j, 0);
                auto found_rib = ribs.find(cur_rib);

                if (found_rib == ribs.end()) {
                    Node left_node = GetNodeById(*i);
                    Node right_node = GetNodeById(*j);
                    int center_id_ = getLastNodesId() + 1;

                    Node center_node(center_id_,
                        std::array<double, 3>{(left_node.coord[0] + right_node.coord[0]) / 2, 
                                              (left_node.coord[1] + right_node.coord[1]) / 2,
                                              (left_node.coord[2] + right_node.coord[2]) / 2},
                        false);

                    cur_rib.UpdateCenter(center_id_);
                    std::move(center_node);
                    el.nodes_id.push_back(center_id_);
                    ribs.insert(cur_rib);

                }
                else {
                    el.nodes_id.push_back(found_rib->center_id);
                }
            }
        }

        el.nodes_id.shrink_to_fit();
    }

    for (auto& it : boundary_finite_elements) {
        std::vector<int> cur_nodes(it.nodes_id);

        for (auto i = cur_nodes.begin(); i < cur_nodes.end() - 1; ++i) {
            for (auto j = i + 1; j < cur_nodes.end(); ++j) {
                Rib cur_rib(*i, *j, 0);
                auto found_rib = ribs.find(cur_rib); 

                if (found_rib != ribs.end()) {
                    it.nodes_id.push_back(found_rib->center_id);
                }
            }
        }

        it.nodes_id.shrink_to_fit();
    }

    boundary_finite_elements.shrink_to_fit();
    nodes.shrink_to_fit();
}

Node MeshLoader::GetNodeById(int n_id) {
    return nodes[n_id];
}

std::vector<Node>& MeshLoader::Nodes() {
    return nodes;
}

std::vector<Node> MeshLoader::VertexNodesBySurfaceId(int s_id) {
    std::set<Node> ans;
    auto cur_iter = boundary_finite_elements.begin();
    auto last_iter = boundary_finite_elements.end();

    while (cur_iter != last_iter) {
        cur_iter = std::find_if(cur_iter, last_iter, [s_id](const BoundaryFiniteElement& s) {
            return s.rib_id == s_id;
            });

        if (cur_iter != last_iter) {
            for (const auto& cur : cur_iter->nodes_id) {
                ans.insert(GetNodeById(s_id));
            }

            ++cur_iter;
        }
    }

    return std::vector<Node>(ans.begin(), ans.end());
}

std::vector<FiniteElement>& MeshLoader::FiniteElements() {
    return finite_elements;
}


std::vector<FiniteElement> MeshLoader::FiniteElementsByThreeId(int id_1, int id_2, int id_3) {
    std::vector<FiniteElement> ans;
    auto cur_iter = finite_elements.begin();
    auto last_iter = finite_elements.end();

    while (cur_iter != last_iter) {
        cur_iter = std::find_if(cur_iter, last_iter, [id_1, id_2, id_3](FiniteElement& cur_elem) {
            return std::find(cur_elem.nodes_id.begin(), cur_elem.nodes_id.end(), id_1) != cur_elem.nodes_id.end()
                && std::find(cur_elem.nodes_id.begin(), cur_elem.nodes_id.end(), id_2) != cur_elem.nodes_id.end()
                && std::find(cur_elem.nodes_id.begin(), cur_elem.nodes_id.end(), id_3) != cur_elem.nodes_id.end();
            });

        if (cur_iter != last_iter) {
            ans.push_back(*cur_iter);
            ++cur_iter;
        }
    }

    return ans;
}

std::vector<FiniteElement> MeshLoader::FiniteElementsByRib(int id_1, int id_2) {
    std::vector<FiniteElement> ans;
    auto cur_iter = finite_elements.begin();
    auto last_iter = finite_elements.end();

    while (cur_iter != last_iter) {
        cur_iter = std::find_if(cur_iter, last_iter, [id_1, id_2](FiniteElement& cur_elem) {
            return std::find(cur_elem.nodes_id.begin(), cur_elem.nodes_id.end(), id_1) != cur_elem.nodes_id.end()
                && std::find(cur_elem.nodes_id.begin(), cur_elem.nodes_id.end(), id_2) != cur_elem.nodes_id.end();
            });

        if (cur_iter != last_iter) {
            ans.push_back(*cur_iter);
            ++cur_iter;
        }
    }

    return ans;
}

std::vector<FiniteElement> MeshLoader::FiniteElementsByMaterialId(int m_id) {
    std::list<FiniteElement> ans;
    auto elem_by_material = finite_elements.begin();
    auto last_elem = finite_elements.end();

    while (elem_by_material != last_elem) {
        elem_by_material = std::find_if(elem_by_material, last_elem, [m_id](FiniteElement& e) {
            return e.material_id == m_id;
            });

        if (elem_by_material != last_elem) {
            ans.push_back(*elem_by_material);
            ++elem_by_material;
        }
    }

    return std::vector<FiniteElement>(ans.begin(), ans.end());
}

std::vector<BoundaryFiniteElement>& MeshLoader::BoundaryFiniteElements() {
    return boundary_finite_elements;
}

std::vector<BoundaryFiniteElement> MeshLoader::BoundaryFiniteElementsById(int s_id) {
    std::list<BoundaryFiniteElement> ans;
    auto surface_by_id = boundary_finite_elements.begin();
    auto last_elem = boundary_finite_elements.end();

    while (surface_by_id != last_elem) {
        surface_by_id = std::find_if(surface_by_id, last_elem, [s_id](BoundaryFiniteElement& s) {
            return s.elem_id == s_id;
            });

        if (surface_by_id != last_elem) {
            ans.push_back(*surface_by_id);
            ++surface_by_id;
        }
    }

    return std::vector<BoundaryFiniteElement>(ans.begin(), ans.end());
}

std::vector<std::vector<Node>> MeshLoader::NodesNeighbors() {
    std::vector<std::set<int>> ans_int(nodes.size() + 1);

    for (const auto& elem : finite_elements) {
        for (auto node_iter : elem.nodes_id) {
            for (auto node_iter2 : elem.nodes_id) {
                ans_int[node_iter].insert(node_iter2);
            }
            ans_int[node_iter].erase(node_iter);
        }
    }

    std::vector<std::vector<Node>> ans(nodes.size() + 1);
    for (int i = 1; i < nodes.size() + 1; ++i) {
        std::vector<Node> tmp;
        tmp.reserve(ans_int[i].size());

        for (auto& set : ans_int[i]) {
            tmp.push_back(GetNodeById(set));
        }

        ans[i] = std::move(tmp);
    }

    return ans;
}
